#!/usr/bin/env python
import unittest
import sys
import os
from pathlib import Path
import json
import pandas as pd
import numpy as np

import emod_api.campaign as camp
import emodpy_tbhiv.demographics.TBHIVDemographics as Demographics

parent = Path(__file__).resolve().parent
sys.path.append(str(parent))
import manifest


class HIVDemographicsTest(unittest.TestCase):
    schema_path = manifest.schema_file

    # region unittest setup and teardown method
    @classmethod
    def setUpClass(cls):
        camp.schema_path = cls.schema_path
        cls.out_folder = os.path.join(manifest.current_directory, 'output', 'demographics')

    def setUp(self):
        print(f"running {self._testMethodName}:")
        pass

    def tearDown(self):
        print("end of test\n")
        pass
    # endregion

    def default_test(self, demog):
        defaults = demog.raw['Defaults']
        self.assertIn('NodeAttributes', defaults)
        self.assertIn('IndividualAttributes', defaults)
        ia = defaults['IndividualAttributes']
        self.assertIn('RiskDistributionFlag', ia)
        self.assertIn('RiskDistribution1', ia)
        self.assertIn('RiskDistribution2', ia)
        self.assertEqual(ia['RiskDistributionFlag'], 1)
        self.assertEqual(ia['RiskDistribution1'], 1)
        self.assertEqual(ia['RiskDistribution2'], 1)
        self.assertIn('AgeDistributionFlag', ia)
        self.assertIn('HIVCoinfectionDistribution', ia)
        self.assertIn('HIVTBCoinfMortalityDistribution', ia)

    # region unittests
    def test_from_basic_node(self):
        out_filename = os.path.join(self.out_folder, "demographics_basic_node.json")
        lat = 11
        lon = 22
        pop = 999
        name = "test_name"
        forced_id = 101
        demog = Demographics.fromBasicNode(lat=lat, lon=lon, pop=pop, name=name, forced_id=forced_id)
        self.assertTrue(isinstance(demog, Demographics.TBHIVDemographics))
        self.default_test(demog)

        print(f"Writing out file: {out_filename}.")
        demog.generate_file(out_filename)
        self.assertTrue(os.path.isfile(out_filename), msg=f'{out_filename} is not generated.')
        with open(out_filename, 'r') as demo_file:
            demog_json = json.load(demo_file)
        self.assertEqual(demog_json['Nodes'][0]['NodeAttributes']['Latitude'], lat)
        self.assertEqual(demog_json['Nodes'][0]['NodeAttributes']['Longitude'], lon)
        self.assertEqual(demog_json['Nodes'][0]['NodeAttributes']['InitialPopulation'], pop)
        self.assertEqual(demog_json['Nodes'][0]['NodeAttributes']['FacilityName'], name)
        self.assertEqual(demog_json['Nodes'][0]['NodeID'], forced_id)

    def test_from_data(self):  # infer_natural_mortality
        male_input_file = os.path.join(manifest.current_directory, "../data/Malawi_male_mortality.csv")
        female_input_file = os.path.join(manifest.current_directory, "../data/Malawi_female_mortality.csv")
        # These are hardcoded values
        interval_fit = [1970, 1980]
        predict_horizon = 2050
        results_scale_factor = 1.0 / 365.0

        pop = 1e6
        demog = Demographics.fromData(pop=pop, filename_male=male_input_file, filename_female=female_input_file)
        self.assertTrue(isinstance(demog, Demographics.TBHIVDemographics))
        self.default_test(demog)

        male_input = pd.read_csv(male_input_file)
        female_input = pd.read_csv(female_input_file)
        output = demog.raw['Defaults']

        # Check pop is honored
        self.assertEqual(demog.nodes[0].pop, pop)

        # Check population group consistency
        male_distribution = output['IndividualAttributes']['MortalityDistributionMale']
        female_distribution = output['IndividualAttributes']['MortalityDistributionFemale']

        # Check age group
        male_age_groups = male_distribution['NumPopulationGroups'][0]
        self.assertEqual(len(male_distribution['PopulationGroups'][0]), male_age_groups)
        # Get age groups information male from csv file
        expected_male_age_group = np.append(male_input['Age (x)'].unique(), 100).tolist()
        self.assertListEqual(expected_male_age_group[1:], male_distribution['PopulationGroups'][0])

        female_age_groups = female_distribution['NumPopulationGroups'][0]
        self.assertEqual(len(female_distribution['PopulationGroups'][0]), female_age_groups)
        # Get age groups information from female csv file
        expected_female_age_group = np.append(female_input['Age (x)'].unique(), 100).tolist()
        self.assertListEqual(expected_female_age_group[1:], female_distribution['PopulationGroups'][0])

        # Check Year
        male_year_groups = male_distribution['NumPopulationGroups'][1]
        self.assertEqual(len(male_distribution['PopulationGroups'][1]), male_year_groups)
        # Get year groups information male from csv file
        expected_male_year_group = male_input['Ave_Year'].unique().tolist()
        # Up to year = predict_horizon
        expected_male_year_group = [x for x in expected_male_year_group if x < predict_horizon]
        self.assertListEqual(expected_male_year_group, male_distribution['PopulationGroups'][1])

        female_year_groups = female_distribution['NumPopulationGroups'][1]
        self.assertEqual(len(female_distribution['PopulationGroups'][1]), female_year_groups)
        expected_female_year_group = expected_male_year_group
        self.assertListEqual(expected_female_year_group, female_distribution['PopulationGroups'][1])

        # Check prediction horizon is honored
        self.assertLessEqual(max(male_distribution['PopulationGroups'][1]), predict_horizon)
        self.assertLessEqual(max(female_distribution['PopulationGroups'][1]), predict_horizon)

        # Check results scale factor is consistent with parameters
        self.assertEqual(male_distribution['ResultScaleFactor'], results_scale_factor)
        self.assertEqual(female_distribution['ResultScaleFactor'], results_scale_factor)

        # Check result values array length
        self.assertEqual(len(male_distribution['ResultValues']), male_age_groups)
        for m_y in male_distribution['ResultValues']:
            self.assertEqual(len(m_y), male_year_groups)
        self.assertEqual(len(female_distribution['ResultValues']), female_age_groups)
        for f_y in female_distribution['ResultValues']:
            self.assertEqual(len(f_y), female_year_groups)

        # Check result values consistency with reference files
        male_reference = pd.read_csv(os.path.join(self.out_folder, "MaleTrue"))
        female_reference = pd.read_csv(os.path.join(self.out_folder, "FemaleTrue"))
        for i in range(male_age_groups):
            for j in range(male_year_groups):
                male_mortality_rate = male_distribution['ResultValues'][i][j]
                expected_male_mortality_rate = male_reference[male_reference['Age'] == expected_male_age_group[i + 1]][
                    str(expected_male_year_group[j])].iloc[0]
                self.assertAlmostEqual(expected_male_mortality_rate, male_mortality_rate, delta=1e-5,
                                       msg=f"at year {expected_male_year_group[j]} age {expected_male_age_group[i + 1]}"
                                           f", male mortality rate is set to {male_mortality_rate} (please see "
                                           f"male_distribution['ResultValues'][{i}][{j}]), while it's "
                                           f"{expected_male_mortality_rate} in male csv file.\n")

                female_mortality_rate = female_distribution['ResultValues'][i][j]
                expected_female_mortality_rate = female_reference[female_reference['Age'] ==
                                                                  expected_female_age_group[i + 1]][
                    str(expected_female_year_group[j])].iloc[0]
                self.assertAlmostEqual(expected_female_mortality_rate, female_mortality_rate, delta=1e-5,
                                       msg=f"at year {expected_female_year_group[j]} age "
                                           f"{expected_female_age_group[i + 1]},"
                                           f" female mortality rate is set to {female_mortality_rate} (please see "
                                           f"female_distribution['ResultValues'][{i}][{j}]), while it's "
                                           f"{expected_female_mortality_rate} in female csv file.\n")
    # endregion


if __name__ == '__main__':
    unittest.main()

